@@ -5,12 +5,46 @@ import Term
 import Rule
 import TRS
 
+contract :: TRS -> Term -> Maybe (Term, Subst)
+contract [] _ = Nothing
+contract ((l, r) : rules) t
+  | Just sigma <- match l t = Just (r, sigma)
+  | otherwise               = contract rules t
+
+normalize :: TRS -> Term -> Subst -> Term
+normalize _   s@(V _)  sigma  = Term.substitute s sigma
+normalize trs (F f ss) sigma =
+  case contract trs t of
+    Nothing       -> t
+    Just (r, tau) -> normalize trs r tau 
+  where
+    t = F f [ normalize trs si sigma | si <- ss ]
+
+nf :: TRS -> Term -> Term
+nf trs t = normalize trs t []
+
+normal_form :: TRS -> Term -> Bool
+normal_form trs t = reducts trs t == []
+
 -- { t | s ->_R^i t for some i <= k }
 at_most_k_steps :: Int -> TRS -> Term -> [Term]
 at_most_k_steps 0 _ s = [ s ]
 at_most_k_steps k trs s =
   nub (s : [ u | t <- reducts trs s, u <- at_most_k_steps (k-1) trs t ])
 
+k_step_reducts' :: TRS -> [Term] ->  [(Int, Term)] -> [Term]
+k_step_reducts' _   _  [] = []
+k_step_reducts' trs ss ((k, t) : pairs)
+  | elem t ss = k_step_reducts' trs ss pairs
+  | k == 0    = t : k_step_reducts' trs (t : ss) pairs
+  | otherwise =
+      case reducts trs t of
+        [] -> t : k_step_reducts' trs (t : ss) pairs
+        us -> k_step_reducts' trs ss (pairs ++ [ (k-1,u) | u <- us ])
+
+k_step_reducts :: Int -> TRS -> Term -> [Term]
+k_step_reducts k trs t = k_step_reducts' trs [] [(k, t)]
+
 at_most_k_walks :: Int -> TRS -> Term -> [(Term, [(Term, Term)])]
 at_most_k_walks 0 _ s = [ (s, []) ]
 at_most_k_walks k trs s =
@@ -18,11 +52,12 @@ at_most_k_walks k trs s =
     where
       ts = at_most_k_walks (k-1) trs s
 
--- { t | s ->_R^i t for some 1 <= i <= k }
-at_least_1_at_most_k_steps :: Int -> TRS -> Term -> [Term]
-at_least_1_at_most_k_steps 0 _   _ = []
-at_least_1_at_most_k_steps k trs t =
-  [ v | u <- reducts trs t, v <- at_most_k_steps (k - 1) trs u ]
+partially_normalize :: Int -> TRS -> Term -> Term
+partially_normalize _ _   s@(V _) = s
+partially_normalize k trs s@(F f ss)
+  | Just u <- find (\v -> normal_form trs v) (at_most_k_steps k trs s) = u
+  | otherwise = F f [ partially_normalize k trs si | si <- ss ]
+
 
 -- { t | s ->_R^epsilon t }
 rootStep :: TRS -> Term -> [Term]
@@ -47,7 +82,7 @@ stepWithRule :: TRS -> Term -> [(Rule, Term)]
 stepWithRule trs s =
   [ ((l, r), Term.replace s t p)
   | p <- Term.positions s,
-    ((l, r), t) <- rootStepWithRule trs (subtermAt s p)]
+    ((l, r), t) <- rootStepWithRule trs (subterm_at s p)]
 
 parallelStepWithRule :: TRS -> Term -> [(TRS, Term)]
 parallelStepWithRule _ s@(V _)    = [([], s)]
@@ -69,18 +104,3 @@ restrictedParallelStep trs xs s@(F f ss) =
        [ (nub (concat ruless), F f ts) 
        | pairs <- sequence [ restrictedParallelStep trs xs si | si <- ss ],
          let (ruless, ts) = unzip pairs ])
-
--- { t | s -o->_R t }
-multistep :: TRS -> Term -> [Term]
-multistep _   s@(V _)    = [s]
-multistep trs s@(F f ss) =
-  nub (s : 
-       [ Term.substitute r tau 
-       | (l, r) <- trs, 
-         Just sigma <- [match l s],
-        tau <- multistepSubst trs sigma ] ++
-        [ F f ts | ts <- sequence [ multistep trs si | si <- ss ] ])
-
-multistepSubst :: TRS -> Subst -> [Subst]
-multistepSubst trs sigma =
-  sequence [ [ (x, t) | t <- multistep trs s ] | (x, s) <- sigma ] 
