@@ -5,8 +5,19 @@ import Term
 
 type Rule = (Term, Term)
 
-showRule :: Rule -> String
-showRule (l, r) = show l ++ " -> " ++ show r
+show_rule :: Rule -> String
+show_rule (l, r) = show l ++ " -> " ++ show r
+
+size :: Rule -> Int
+size (l, r) = Term.size l + Term.size r
+
+subset :: Eq a => [a] -> [a] -> Bool
+subset xs ys = all p xs
+  where p x = elem x ys
+
+well_formed :: Rule -> Bool
+well_formed (l@(F _ _), r) = subset (Term.variables r) (Term.variables l)
+well_formed _ = False
 
 -- Var(l -> r) = Var(l) cup Var(r)
 variables :: Rule -> [String]
@@ -16,12 +27,8 @@ variables (l, r) = union (Term.variables l) (Term.variables r)
 functions :: Rule -> [String]
 functions (l, r) = union (Term.functions l) (Term.functions r)
 
-signatureOf :: Rule -> Signature
-signatureOf (l, r) = union (Term.signatureOf l) (Term.signatureOf r) 
-
--- |l -> r| = |l| + |r|
-size :: Rule -> Int
-size (l, r) = Term.size l + Term.size r
+signature_of :: Rule -> Signature
+signature_of (l, r) = union (Term.signature_of l) (Term.signature_of r) 
 
 substitute :: Rule -> Subst -> Rule
 substitute (l, r) sigma =
@@ -41,34 +48,8 @@ variant rule1 rule2 =
   Rule.subsume rule1 rule2 &&
   Rule.subsume rule2 rule1
 
-subset :: Eq a => [a] -> [a] -> Bool
-subset xs ys = all (\x -> elem x ys) xs
+left_linear :: Rule -> Bool
+left_linear (l, _) = Term.linear l
 
-extraVariables :: Rule -> [String]
-extraVariables (l, r) =
-  nub [ x | x <- Term.variables r, not (elem x (Term.variables l)) ]
-
-variableCondition :: Rule -> Bool
-variableCondition (V _, _) = False
-variableCondition (l, r)   = subset (Term.variables r) (Term.variables l)
-
-nonduplicating :: Rule -> Bool
-nonduplicating (l, r) = all p (Term.variables r)
-  where p x = count l x >= count r x
-
-strictlyLengthReducing :: Rule -> Bool
-strictlyLengthReducing (l, r) =
-  Term.size l > Term.size r && nonduplicating (l, r)
-
-collapsing :: Rule -> Bool
-collapsing (_, V _) = True
-collapsing _        = False
-
-linear :: Rule -> Bool
-linear (l, r) = Term.linear l && Term.linear r
-
-leftLinear :: Rule -> Bool
-leftLinear (l, _) = Term.linear l
-
-rightLinear :: Rule -> Bool
-rightLinear (_, r) = Term.linear r
+non_left_linear :: Rule -> Bool
+non_left_linear rule = not (Rule.left_linear rule)
